@@ -42,6 +42,7 @@
 #include "wayland-util.h"
 #include "wayland-private.h"
 #include "wayland-server-core.h"
+#include "wayland-server-private.h"
 #include "wayland-os.h"
 
 /** \cond INTERNAL */
@@ -73,7 +74,7 @@ struct wl_event_loop {
 	struct wl_list idle_list;
 	struct wl_list destroy_list;
 
-	struct wl_signal destroy_signal;
+	struct wl_priv_signal destroy_signal;
 
 	struct wl_timer_heap timers;
 };
@@ -179,7 +180,7 @@ wl_event_loop_add_fd(struct wl_event_loop *loop,
 {
 	struct wl_event_source_fd *source;
 
-	source = malloc(sizeof *source);
+	source = zalloc(sizeof *source);
 	if (source == NULL)
 		return NULL;
 
@@ -340,7 +341,7 @@ wl_timer_heap_reserve(struct wl_timer_heap *timers)
 		new_space = timers->space >= 8 ? timers->space * 2 : 8;
 		n = realloc(timers->data, (size_t)new_space * sizeof(*n));
 		if (!n) {
-			wl_log("Allocation failure when expanding timer list");
+			wl_log("Allocation failure when expanding timer list\n");
 			return -1;
 		}
 		timers->data = n;
@@ -361,7 +362,7 @@ wl_timer_heap_unreserve(struct wl_timer_heap *timers)
 	if (timers->space >= 16 && timers->space >= 4 * timers->count) {
 		n = realloc(timers->data, (size_t)timers->space / 2 * sizeof(*n));
 		if (!n) {
-			wl_log("Reallocation failure when shrinking timer list");
+			wl_log("Reallocation failure when shrinking timer list\n");
 			return;
 		}
 		timers->data = n;
@@ -568,7 +569,7 @@ wl_event_loop_add_timer(struct wl_event_loop *loop,
 	if (wl_timer_heap_ensure_timerfd(&loop->timers) < 0)
 		return NULL;
 
-	source = malloc(sizeof *source);
+	source = zalloc(sizeof *source);
 	if (source == NULL)
 		return NULL;
 
@@ -718,7 +719,7 @@ wl_event_loop_add_signal(struct wl_event_loop *loop,
 	struct wl_event_source_signal *source;
 	sigset_t mask;
 
-	source = malloc(sizeof *source);
+	source = zalloc(sizeof *source);
 	if (source == NULL)
 		return NULL;
 
@@ -775,7 +776,7 @@ wl_event_loop_add_idle(struct wl_event_loop *loop,
 {
 	struct wl_event_source_idle *source;
 
-	source = malloc(sizeof *source);
+	source = zalloc(sizeof *source);
 	if (source == NULL)
 		return NULL;
 
@@ -885,7 +886,7 @@ wl_event_loop_create(void)
 {
 	struct wl_event_loop *loop;
 
-	loop = malloc(sizeof *loop);
+	loop = zalloc(sizeof *loop);
 	if (loop == NULL)
 		return NULL;
 
@@ -898,7 +899,7 @@ wl_event_loop_create(void)
 	wl_list_init(&loop->idle_list);
 	wl_list_init(&loop->destroy_list);
 
-	wl_signal_init(&loop->destroy_signal);
+	wl_priv_signal_init(&loop->destroy_signal);
 
 	wl_timer_heap_init(&loop->timers, loop);
 
@@ -921,7 +922,7 @@ wl_event_loop_create(void)
 WL_EXPORT void
 wl_event_loop_destroy(struct wl_event_loop *loop)
 {
-	wl_signal_emit(&loop->destroy_signal, loop);
+	wl_priv_signal_final_emit(&loop->destroy_signal, loop);
 
 	wl_event_loop_process_destroy_list(loop);
 	wl_timer_heap_release(&loop->timers);
@@ -942,8 +943,8 @@ post_dispatch_check(struct wl_event_loop *loop)
 
 		dispatch_result = source->interface->dispatch(source, &ep);
 		if (dispatch_result < 0) {
-			wl_log("Source dispatch function returned negative value!");
-			wl_log("This would previously accidentally suppress a follow-up dispatch");
+			wl_log("Source dispatch function returned negative value!\n");
+			wl_log("This would previously accidentally suppress a follow-up dispatch\n");
 		}
 		needs_recheck |= dispatch_result != 0;
 	}
@@ -971,6 +972,57 @@ wl_event_loop_dispatch_idle(struct wl_event_loop *loop)
 	}
 }
 
+static int
+timespec_to_ms(struct timespec value)
+{
+	return (value.tv_sec * 1000) + (value.tv_nsec / 1000000);
+}
+
+static struct timespec
+ms_to_timespec(int ms)
+{
+	struct timespec val;
+	val.tv_sec = ms / 1000;
+	val.tv_nsec = (ms % 1000) * 1000000;
+	return val;
+}
+
+static struct timespec
+timespec_normalize(struct timespec value)
+{
+	struct timespec result = value;
+
+	while (result.tv_nsec >= 1000000000) {
+		result.tv_nsec -= 1000000000;
+		result.tv_sec++;
+	}
+
+	while (result.tv_nsec < 0) {
+		result.tv_nsec += 1000000000;
+		result.tv_sec--;
+	}
+
+	return result;
+}
+
+static struct timespec
+timespec_add(struct timespec a, struct timespec b)
+{
+	struct timespec result;
+	result.tv_sec = a.tv_sec + b.tv_sec;
+	result.tv_nsec = a.tv_nsec + b.tv_nsec;
+	return timespec_normalize(result);
+}
+
+static struct timespec
+timespec_sub(struct timespec a, struct timespec b)
+{
+	struct timespec result;
+	result.tv_sec = a.tv_sec - b.tv_sec;
+	result.tv_nsec = a.tv_nsec - b.tv_nsec;
+	return timespec_normalize(result);
+}
+
 /** Wait for events and dispatch them
  *
  * \param loop The event loop whose sources to wait for.
@@ -998,17 +1050,43 @@ wl_event_loop_dispatch(struct wl_event_loop *loop, int timeout)
 	struct wl_event_source *source;
 	int i, count;
 	bool has_timers = false;
+	bool use_timeout = timeout > 0;
+	struct timespec now, end;
 
 	wl_event_loop_dispatch_idle(loop);
 
-	count = epoll_wait(loop->epoll_fd, ep, ARRAY_LENGTH(ep), timeout);
+	if (use_timeout) {
+		clock_gettime(CLOCK_MONOTONIC, &now);
+		end = timespec_add(now, ms_to_timespec(timeout));
+	}
+
+	while (true) {
+		count = epoll_wait(loop->epoll_fd, ep, ARRAY_LENGTH(ep), timeout);
+		if (count >= 0)
+			break; /* have events or timeout */
+		else if (count < 0 && errno != EINTR && errno != EAGAIN)
+			break; /* have error */
+
+		if (use_timeout) {
+			clock_gettime(CLOCK_MONOTONIC, &now);
+			timeout = timespec_to_ms(timespec_sub(end, now));
+			if (timeout <= 0) {
+				/* too late */
+				count = 0;
+				break;
+			}
+		}
+	}
+
 	if (count < 0)
 		return -1;
 
 	for (i = 0; i < count; i++) {
 		source = ep[i].data.ptr;
-		if (source == &loop->timers.base)
+		if (source == &loop->timers.base) {
 			has_timers = true;
+			break;
+		}
 	}
 
 	if (has_timers) {
@@ -1070,7 +1148,7 @@ WL_EXPORT void
 wl_event_loop_add_destroy_listener(struct wl_event_loop *loop,
 				   struct wl_listener *listener)
 {
-	wl_signal_add(&loop->destroy_signal, listener);
+	wl_priv_signal_add(&loop->destroy_signal, listener);
 }
 
 /** Get the listener struct for the specified callback
@@ -1086,5 +1164,5 @@ WL_EXPORT struct wl_listener *
 wl_event_loop_get_destroy_listener(struct wl_event_loop *loop,
 				   wl_notify_func_t notify)
 {
-	return wl_signal_get(&loop->destroy_signal, notify);
+	return wl_priv_signal_get(&loop->destroy_signal, notify);
 }
